from sklearn.datasets import fetch_lfw_people
import argparse
from tqdm import tqdm
import os
import time
import shutil
import torch
import numpy as np
import torchvision
from torchmetrics.image.fid import FrechetInceptionDistance
import matplotlib.pyplot as plt
from torch import autograd

"""
Changes to WGAN

1. Use gradient penalty instead of weight clipping -> use instance norm in critic
2. Scale images to [-1; 1] and use tanh in generator
3. Use Frechet Inception Distance for evaluation
4. Replay buffer to train C on old generations

Namespace(run_path='wgan-gp', num_epochs=1000, batch_size=64, learning_rate=0.0001, betas=(0.0, 0.9), dropout=0.0, z_size=100, critic_iters=5, gp_weight=10, fid_n_samples=10000, calculate_fid_every=1, buffer_size=4, gpu_buffer=True, debug=False)
"""



def parse_args():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-run_path', default='wgan-gp', type=str)
    parser.add_argument('-num_epochs', default=1000, type=int)
    parser.add_argument('-batch_size', default=64, type=int)
    parser.add_argument('-learning_rate', default=1e-4, type=float)
    parser.add_argument('-betas', default=(0.0, 0.9), type=tuple)
    parser.add_argument('-dropout', default=0.0, type=float)
    parser.add_argument('-z_size', default=100, type=int)
    parser.add_argument('-critic_iters', default=5, type=int)
    parser.add_argument('-gp_weight', default=10, type=float, help='Gradient penalty weight')
    parser.add_argument('-fid_n_samples', default=10000, type=int)
    parser.add_argument('-calculate_fid_every', default=1, type=int)
    parser.add_argument('-buffer_size', default=4, type=float,
                        help='Size of replay buffer as a fraction of dataset size')
    parser.add_argument('-gpu_buffer', action='store_true',
                        help='Store replay buffer on GPU (requires 24GB vram for default buffer size')
    parser.add_argument('-debug', action='store_true')

    args, _ = parser.parse_known_args()

    return args


class DatasetLFW(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()

        # 96x96 src images scaled to 64x64
        data_dict = fetch_lfw_people(
            data_home='../data',
            color=True,
            resize=2 / 3,
            slice_=(slice(95, 191, None), slice(77, 173, None))
        )

        np_x = np.transpose(data_dict.images, (0, 3, 1, 2))
        np_y = data_dict.target

        self.data = torch.FloatTensor(np_x) * 2 - 1
        self.labels = torch.LongTensor(np_y)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]


class CBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, stride=2, norm='instance', dropout=0.):
        super().__init__()

        assert norm in ['instance', 'batch', None], 'Invalid norm type'

        self.conv = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )

        if norm == 'instance':
            self.norm = torch.nn.InstanceNorm2d(out_channels)
        elif norm == 'batch':
            self.norm = torch.nn.BatchNorm2d(out_channels)
        else:
            self.norm = torch.nn.Identity()

        self.act = torch.nn.LeakyReLU(0.2, inplace=True)
        self.dropout = torch.nn.Dropout2d(dropout)

    def forward(self, x):
        x = self.conv(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.dropout(x)
        return x


class ModelC(torch.nn.Module):
    def __init__(self, dropout=0.):
        super().__init__()

        self.encoder_blocks = torch.nn.ModuleList([
            CBlock(3, 64, stride=2, dropout=dropout),  # 64->32
            CBlock(64, 64, stride=1, dropout=dropout),  # 32->32
            CBlock(64, 128, stride=2, dropout=dropout),  # 32->16
            CBlock(128, 128, stride=1, dropout=dropout),  # 16->16
            CBlock(128, 256, stride=2, dropout=dropout),  # 16->8
            CBlock(256, 256, stride=1, dropout=dropout),  # 8->8
            CBlock(256, 512, stride=2, dropout=dropout),  # 8->4
            CBlock(512, 512, stride=1, dropout=dropout),  # 4->4
        ])

        self.conv = torch.nn.Conv2d(
            in_channels=512,
            out_channels=1,
            kernel_size=4,
            stride=1,
            padding=0,
            bias=False,
        )
        self._initialize_weights()

    def enable_dropout(self, mode=True):
        for block in self.encoder_blocks:
            block.dropout.train(mode)

    def forward(self, x):
        for block in self.encoder_blocks:
            x = block(x)
        x = self.conv(x)
        return x

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.normal_(m.weight, 0.0, 0.02)
                if m.bias is not None:
                    torch.nn.init.constant_(m.bias, 0)
            elif isinstance(m, torch.nn.BatchNorm2d):
                torch.nn.init.normal_(m.weight, 1.0, 0.02)
                if m.bias is not None:
                    torch.nn.init.constant_(m.bias, 0)


class GBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, act='leakyrelu', dropout=0.):
        super().__init__()

        self.upsample = torch.nn.UpsamplingBilinear2d(scale_factor=2)
        self.conv = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )

        self.norm = torch.nn.BatchNorm2d(out_channels)

        if act == 'leakyrelu':
            self.act = torch.nn.LeakyReLU(0.2, inplace=True)
        elif act == 'tanh':
            self.act = torch.nn.Tanh()

        self.dropout = torch.nn.Dropout2d(dropout)

    def forward(self, x):
        x = self.upsample(x)
        x = self.conv(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.dropout(x)
        return x


class ModelG(torch.nn.Module):
    def __init__(self, z_size=100, dropout=0.):
        super().__init__()

        self.fc = torch.nn.Linear(z_size, 4 * 4 * 512)

        self.decoder_blocks = torch.nn.ModuleList([
            GBlock(512, 256, act='leakyrelu', dropout=dropout),  # 4->8
            GBlock(256, 128, act='leakyrelu', dropout=dropout),  # 8->16
            GBlock(128, 64, act='leakyrelu', dropout=dropout),  # 16->32
            GBlock(64, 3, act='tanh'),  # 32->64
        ])

        self._initialize_weights()

    def enable_dropout(self, mode=True):
        for block in self.decoder_blocks:
            block.dropout.train(mode)

    def forward(self, x):
        x = self.fc(x)
        x = x.view(x.size(0), 512, 4, 4)
        for block in self.decoder_blocks:
            x = block(x)
        return x

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.normal_(m.weight, 0.0, 0.02)
                if m.bias is not None:
                    torch.nn.init.constant_(m.bias, 0)
            elif isinstance(m, torch.nn.BatchNorm2d):
                torch.nn.init.normal_(m.weight, 1.0, 0.02)
                if m.bias is not None:
                    torch.nn.init.constant_(m.bias, 0)


class BatchReplayBuffer:
    """"
    As featured in "Learning from Simulated and Unsupervised Images through Adversarial Training" (https://arxiv.org/pdf/1612.07828.pdf)

    B - size of buffer
    b - training mini batch size

    At each iteration of discriminator training:
    1. Sample b/2 images from the buffer
    2. Sample b/2 images from the current generator
    3. Update discriminator using both samples

    After each iteration:
    Randomly replace b/2 images in the buffer with newly generated ones
    """

    def __init__(self, max_size, device):
        self.max_size = max_size
        self.device = device
        self.data: torch.Tensor = None

    def push(self, batch: torch.Tensor):
        if self.data is None:
            self.data = batch
            return

        n_free = self.max_size - self.data.size(0)

        if n_free > 0:
            self._replace_batch(batch[n_free:])
            self.data = torch.cat((self.data, batch[:n_free]))
        else:
            self._replace_batch(batch)

    def sample(self, batch_size: int) -> torch.Tensor:
        idx = self._random_choice_idx(batch_size)
        return self.data[idx]

    def _replace_batch(self, batch: torch.Tensor):
        idx = self._random_choice_idx(batch.size(0))
        self.data.index_copy_(0, torch.tensor(idx, device=self.device), batch)

    def _random_choice_idx(self, batch_size: int) -> torch.Tensor:
        assert self.data.size(0) >= batch_size, "Attempted to sample too many elements"
        chosen_indices = torch.randint(0, self.data.size(0), (batch_size,), device=self.device)
        return chosen_indices

    def __len__(self):
        if self.data is None:
            return 0
        return len(self.data)


def compute_gp(netC, real_data, fake_data):
    """
    Gradient Penalty implementation from
    https://towardsdatascience.com/demystified-wasserstein-gan-with-gradient-penalty-ba5e9b905ead#:~:text=The%20idea%20of%20Gradient%20Penalty,unit%20norm%20(Statement%201).&text=In%20Eq.,sum%20is%20the%20gradient%20penalty.
    """
    batch_size = real_data.size(0)
    # Sample Epsilon from uniform distribution
    eps = torch.rand(batch_size, 1, 1, 1).to(real_data.device)
    eps = eps.expand_as(real_data)

    # Interpolation between real data and fake data.
    interpolation = eps * real_data + (1 - eps) * fake_data
    interpolation.requires_grad_(True)

    # get logits for interpolated images
    interp_logits = netC(interpolation)
    grad_outputs = torch.ones_like(interp_logits)

    # Compute Gradients
    gradients = autograd.grad(
        outputs=interp_logits,
        inputs=interpolation,
        grad_outputs=grad_outputs,
        create_graph=True,
        retain_graph=True,
    )[0]

    # Compute and return Gradient Norm
    gradients = gradients.view(batch_size, -1)
    grad_norm = gradients.norm(2, 1)
    return torch.mean((grad_norm - 1) ** 2)


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def main():
    args = parse_args()

    RUN_PATH = args.run_path
    EPOCHS = args.num_epochs
    BATCH_SIZE = args.batch_size
    assert BATCH_SIZE % 2 == 0, 'Batch size must be even for replay buffer to work'
    LEARNING_RATE = args.learning_rate
    BETAS = args.betas
    Z_SIZE = args.z_size
    DROPOUT = args.dropout
    CRITIC_ITERS = args.critic_iters
    GP_WEIGHT = args.gp_weight
    FID_N_SAMPLES = args.fid_n_samples
    CALCULATE_FID_EVERY = args.calculate_fid_every
    BUFFER_SIZE = args.buffer_size
    DEVICE = 'cuda'
    BUFFER_DEVICE = 'cuda' if args.gpu_buffer else 'cpu'
    IS_DEBUG = args.debug

    plt.rcParams["figure.figsize"] = (15, 7)
    plt.style.use('dark_background')
    if not torch.cuda.is_available() or IS_DEBUG:
        DEVICE = 'cpu'
        BATCH_SIZE = 16

    if len(RUN_PATH):
        RUN_PATH = time.strftime(f'runs/{RUN_PATH}-%Y-%m-%d-%H-%M-%S')

    if os.path.exists(RUN_PATH):
        shutil.rmtree(RUN_PATH)

    txt_to_save = {}
    args_txt = str(args)
    txt_to_save['args.txt'] = args_txt

    dataset = DatasetLFW()

    data_loader = torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=BATCH_SIZE,
        shuffle=True,
        drop_last=True,
        num_workers=(4 if not IS_DEBUG else 0),
    )

    model_C = ModelC(dropout=DROPOUT).to(DEVICE)
    model_G = ModelG(dropout=DROPOUT).to(DEVICE)

    c_params = count_parameters(model_C)
    g_params = count_parameters(model_G)

    print(f'Number of parameters in C: {c_params}')
    print(f'Number of parameters in G: {g_params}')
    print(f'c/g: {c_params / g_params: .2f}')

    txt_to_save['model_C.txt'] = str(model_C) + f'\n\nNumber of parameters: {c_params}'
    txt_to_save['model_G.txt'] = str(model_G) + f'\n\nNumber of parameters: {g_params}'

    optimizer_G = torch.optim.Adam(
        model_G.parameters(), lr=LEARNING_RATE, betas=BETAS)

    optimizer_C = torch.optim.Adam(
        model_C.parameters(), lr=LEARNING_RATE, betas=BETAS)

    fid = FrechetInceptionDistance(feature=192).to(DEVICE)

    buffer_size = int(BUFFER_SIZE * len(dataset))
    batch_replay_buffer = BatchReplayBuffer(max_size=buffer_size, device=BUFFER_DEVICE)

    metrics = {}
    for stage in ['train']:
        for metric in ['loss_g', 'loss_c', 'fid']:
            metrics[f'{stage}_{metric}'] = []

    z_dist = torch.distributions.normal.Normal(0, 1)

    best_epoch = 0
    for epoch in range(1, EPOCHS):
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        for i, (x_real, _) in enumerate(tqdm(data_loader, desc=stage)):
            x_real = x_real.to(DEVICE)

            optimizer_C.zero_grad()

            y_pred_real = model_C.forward(x_real).view(-1)

            z = z_dist.sample((x_real.size(0) // 2, Z_SIZE)).to(DEVICE)

            # Disable dropout in G when training C for more focused critique of current G flaws
            # To avoid overfitting to current G, replay buffer is used
            # i.e. instead of adding noise, train on previous G states
            model_G.enable_dropout(False)
            x_fake = model_G.forward(z)
            model_G.enable_dropout(True)

            # Sample equaly from replay buffer and new generations
            if len(batch_replay_buffer) > 0:
                buffer_x_fake = batch_replay_buffer.sample(x_real.size(0) // 2).to(DEVICE)
                batch_replay_buffer.push(x_fake.detach().to(BUFFER_DEVICE))
                x_fake = torch.cat((x_fake, buffer_x_fake), dim=0)
            else:
                batch_replay_buffer.push(x_fake.detach().to(BUFFER_DEVICE))
                z = z_dist.sample((x_real.size(0) // 2, Z_SIZE)).to(DEVICE)
                x_fake = torch.cat((x_fake, model_G.forward(z)), dim=0)

            y_pred_fake = model_C.forward(x_fake.detach()).view(-1)

            gp = compute_gp(model_C, x_real, x_fake.detach())
            loss_C = torch.mean(y_pred_fake) - torch.mean(y_pred_real) + GP_WEIGHT * gp
            loss_C.backward()

            optimizer_C.step()

            metrics_epoch[f'{stage}_loss_c'].append(loss_C.cpu().item())

            if i % CRITIC_ITERS == 0:
                z = z_dist.sample((x_real.size(0), Z_SIZE)).to(DEVICE)
                x_fake = model_G.forward(z)
                optimizer_G.zero_grad()

                # Disable dropout in C when training G to get better gradients
                model_C.enable_dropout(False)
                y_pred_fake = model_C.forward(x_fake).view(-1)
                model_C.enable_dropout(True)

                loss_G = -torch.mean(y_pred_fake)
                loss_G.backward()
                optimizer_G.step()

                metrics_epoch[f'{stage}_loss_g'].append(loss_G.cpu().item())

        ### Evaluation ####################################
        model_G.eval()

        # Compute FID
        if epoch % CALCULATE_FID_EVERY == 1 or CALCULATE_FID_EVERY == 1:
            fid.reset()
            n_batches = FID_N_SAMPLES // BATCH_SIZE + 1
            for i, (x_real, _) in enumerate(data_loader):
                x_real = x_real.to(DEVICE)
                z = z_dist.sample((x_real.size(0), Z_SIZE)).to(DEVICE)
                x_fake = model_G.forward(z)

                x_real = (((x_real + 1) / 2) * 255).to(torch.uint8)
                x_fake = (((x_fake + 1) / 2) * 255).to(torch.uint8)

                fid.update(x_fake, real=False)
                fid.update(x_real, real=True)

                if i >= n_batches:
                    break

            metrics_epoch['train_fid'] = [fid.compute().cpu().item()]
        else:
            metrics_epoch['train_fid'] = [metrics['train_fid'][-1]]

        # Generate samples without dropout
        z = z_dist.sample((16, Z_SIZE)).to(DEVICE)
        x_fake = model_G.forward(z)

        model_G.train()
        ##################################################

        metrics_strs = []
        for key in metrics_epoch.keys():
            if 'fake' in key or 'real' in key:
                continue
            value = 0
            if len(metrics_epoch[key]):
                value = np.mean(metrics_epoch[key])
            metrics[key].append(value)
            metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

        plt.clf()
        plt.subplot(121)  # row col idx
        plts = []
        c = 0
        for key, value in metrics.items():
            plts += plt.plot(value, f'C{c}', label=key)
            ax = plt.twinx()
            c += 1
        plt.legend(plts, [it.get_label() for it in plts])

        plt.subplot(122)  # row col idx
        grid_img = torchvision.utils.make_grid(
            (x_fake.detach().cpu() + 1) / 2,
            padding=10,
            scale_each=True,
            nrow=4
        )
        plt.imshow(grid_img.permute(1, 2, 0))

        if len(RUN_PATH) and not os.path.exists(RUN_PATH):
            os.makedirs(RUN_PATH)
            for filename, txt in txt_to_save.items():
                with open(f'{RUN_PATH}/{filename}', 'w') as f:
                    f.write(txt)

        if len(RUN_PATH) == 0:
            plt.show()
        else:
            current_file = f'{RUN_PATH}/plt-{epoch}.png'
            plt.savefig(current_file)
            shutil.copy(current_file, f'{RUN_PATH}/plt-latest.png')

            if epoch == 1 or metrics['train_fid'][-1] < metrics['train_fid'][best_epoch - 1]:
                torch.save(model_G.state_dict(), f'{RUN_PATH}/model_G-{epoch}.pt')
                torch.save(model_C.state_dict(), f'{RUN_PATH}/model_C-{epoch}.pt')
                if os.path.exists(f'{RUN_PATH}/model_G-{best_epoch}.pt'):
                    os.remove(f'{RUN_PATH}/model_G-{best_epoch}.pt')
                if os.path.exists(f'{RUN_PATH}/model_C-{best_epoch}.pt'):
                    os.remove(f'{RUN_PATH}/model_C-{best_epoch}.pt')
                best_epoch = epoch


if __name__ == '__main__':
    main()