"""
CHANGES:
- Added support for non-square images
- Added dropouts to the generator https://arxiv.org/pdf/1611.07004v1.pdf
- Replaced weight clipping with gradient penalty (Improved Training of Wasserstein GANs, 2017)
    - Changed critic BatchNorm2D to LayerNorm
    - Added gradient penalty to the loss
- Changed optimizer to Adam and changed hyperparameters according to Improved Training of Wasserstein GANs, 2017

Pievienoju labāko rezultātu līdz apnika gaidīt :)
"""

import argparse  # pip3 install argparse
from copy import copy

import hashlib
import os
import pickle
import time
import shutil
from tqdm import tqdm  # pip install tqdm
import numpy as np
import torchvision
import torch
import torch.distributed
import torch.utils.datao
import sklearn.datasets

import matplotlib.pyplot as plt

plt.rcParams["figure.figsize"] = (15, 7)
plt.style.use('dark_background')

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-run_path', default='', type=str)

parser.add_argument('-num_epochs', default=100, type=int)
parser.add_argument('-batch_size', default=64, type=int)
parser.add_argument('-classes_count', default=0, type=int)
parser.add_argument('-samples_per_class', default=0, type=int)

parser.add_argument('-learning_rate', default=1e-4, type=float)
parser.add_argument('-z_size', default=128, type=int)

parser.add_argument('-is_debug', default=False, type=lambda x: (str(x).lower() == 'true'))

args, _ = parser.parse_known_args()

RUN_PATH = args.run_path
BATCH_SIZE = args.batch_size
EPOCHS = args.num_epochs
LEARNING_RATE = args.learning_rate
Z_SIZE = args.z_size
DEVICE = 'cuda'
MAX_LEN = args.samples_per_class
MAX_CLASSES = args.classes_count  # 0 = include all
IS_DEBUG = args.is_debug

# W, H; selected such that they are divisible by 4, but still close to the default (47x62)
INPUT_SIZE = np.array([48, 60])
IMG_SCALE = 0.25
PENALTY_COEF = 10

if not torch.cuda.is_available() or IS_DEBUG:
    DEVICE = 'cpu'
    BATCH_SIZE = 66

if len(RUN_PATH):
    RUN_PATH = f'{int(time.time())}_{RUN_PATH}'
    if os.path.exists(RUN_PATH):
        shutil.rmtree(RUN_PATH)
    os.makedirs(RUN_PATH)


class DatasetEMNIST(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()  # 62 classes
        self.data = torchvision.datasets.EMNIST(
            root='../data',
            split='balanced',
            train=(MAX_LEN == 0),
            download=True
        )
        class_to_idx = self.data.class_to_idx
        idx_to_class = dict((value, key) for key, value in class_to_idx.items())
        self.labels = [idx_to_class[idx] for idx in range(len(idx_to_class))]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor
        pil_x, y_idx = self.data[idx]
        np_x = np.transpose(np.array(pil_x)).astype(np.float32)
        np_x = np.expand_dims(np_x, axis=0) / 255.0  # (1, W, H) => (1, 28, 28)
        return np_x, y_idx


class DatasetFashionMNIST(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()  # 62 classes
        self.data = torchvision.datasets.FashionMNIST(
            root='../data',
            # split='balanced',
            train=(MAX_LEN == 0),
            download=True
        )
        class_to_idx = self.data.class_to_idx
        idx_to_class = dict((value, key) for key, value in class_to_idx.items())
        self.labels = [idx_to_class[idx] for idx in range(len(idx_to_class))]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor
        pil_x, y_idx = self.data[idx]
        np_x = np.array(pil_x).astype(np.float32)
        np_x = np.expand_dims(np_x, axis=0) / 255.0  # (1, W, H) => (1, 28, 28)
        return np_x, y_idx


class DatasetLFW(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()

        # Preslice the images to fit requirements
        target_img_size = INPUT_SIZE
        default_w_slice = slice(70, 195)
        default_h_slice = slice(78, 172)

        pre_scale_size = np.round(target_img_size / IMG_SCALE)
        default_w = default_w_slice.stop - default_w_slice.start
        default_h = default_h_slice.stop - default_h_slice.start

        delta_w = pre_scale_size[0] - default_w
        delta_h = pre_scale_size[1] - default_h

        w_slice = slice(
            round(default_w_slice.start - delta_w / 2),
            round(default_w_slice.stop + delta_w / 2)
        )

        h_slice = slice(
            round(default_h_slice.start - delta_h / 2),
            round(default_h_slice.stop + delta_h / 2)
        )

        self.data = sklearn.datasets.fetch_lfw_people(
            data_home="../data",
            download_if_missing=True,
            resize=IMG_SCALE,
            slice_=(h_slice, w_slice),
        )

        self.labels = list(self.data.target_names)
        self.images = self.data.images

    def __len__(self):
        return len(self.data.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor

        # Why does the documentation say that the data is 0-255?????
        # That one unpaid intern documenting sklearn cost me hours

        x = self.images[idx]
        x = np.expand_dims(x, axis=0)
        y_idx = self.data.target[idx]
        return x, y_idx


class ModelD(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.input_size = np.array([INPUT_SIZE[1], INPUT_SIZE[0]])

        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=1, out_channels=8, kernel_size=3, stride=1, padding=1),  # (B, 1, w, h)
            torch.nn.LayerNorm(normalized_shape=self.input_size),
            torch.nn.LeakyReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),  # (B, 8, w/2, h/2)

            torch.nn.Conv2d(in_channels=8, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.LayerNorm(normalized_shape=self.input_size // 2),
            torch.nn.LeakyReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),  # (B, 32, w/4, h/4)

            torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.LayerNorm(normalized_shape=self.input_size // 4),
            torch.nn.LeakyReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),  # (B, 8, w/8, h/8)

            torch.nn.AdaptiveMaxPool2d((1, 1))  # (B, 64, 1, 1)
        )

        self.fc = torch.nn.Linear(in_features=64, out_features=1)

    def forward(self, x):
        out = self.encoder(x)
        out = self.fc(out.squeeze())
        # out = torch.sigmoid(out)
        return out


class ModelG(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.init_size = INPUT_SIZE // 4
        self.fc = torch.nn.Linear(in_features=Z_SIZE, out_features=128 * self.init_size[0] * self.init_size[1])

        self.decoder = torch.nn.Sequential(
            torch.nn.BatchNorm2d(num_features=128),  # (B, 128, 1, 1)
            torch.nn.Upsample(scale_factor=2),  # (B, 128, 2, 2)

            torch.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=128),
            torch.nn.Dropout2d(),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),

            torch.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=128),
            torch.nn.Dropout2d(),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),

            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(in_channels=128, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=64),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),

            torch.nn.Conv2d(in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),

            torch.nn.Conv2d(in_channels=32, out_channels=1, kernel_size=3, stride=1, padding=1),
            torch.nn.Sigmoid()
        )

    def forward(self, z):
        out = self.fc(z)
        out = out.view(-1, 128, self.init_size[1], self.init_size[0])
        out = self.decoder(out)

        return out


# dataset_full = DatasetEMNIST()
# dataset_full = DatasetFashionMNIST()
dataset_full = DatasetLFW()

np.random.seed(2)
labels_train = copy(dataset_full.labels)
np.random.shuffle(labels_train)  # , random=np.random.random

if MAX_CLASSES != 0:
    labels_train = labels_train[:MAX_CLASSES]

np.random.seed(int(time.time()))

print(f"train labels: {labels_train}")

idx_train = []
str_args_for_hasing = [str(it) for it in [MAX_LEN, MAX_CLASSES] + labels_train]
hash_args = hashlib.md5((''.join(str_args_for_hasing)).encode()).hexdigest()
path_cache = f'../data/{hash_args}_gan.pkl'
if os.path.exists(path_cache):
    print('loading from cache')
    with open(path_cache, 'rb') as fp:
        idx_train = pickle.load(fp)
else:
    labels_count = dict((key, 0) for key in dataset_full.labels)
    for idx, (x, y_idx) in tqdm(
            enumerate(dataset_full),
            'splitting dataset',
            total=len(dataset_full)
    ):
        label = dataset_full.labels[y_idx]
        if MAX_LEN > 0:
            if labels_count[label] >= MAX_LEN:
                if all(it >= MAX_LEN for it in labels_count.values()):
                    break
                continue
        labels_count[label] += 1
        if label in labels_train:
            idx_train.append(idx)

    with open(path_cache, 'wb') as fp:
        pickle.dump(idx_train, fp)

dataset_train = torch.utils.data.Subset(dataset_full, idx_train)
data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE < 12),
    num_workers=(8 if not IS_DEBUG else 0)
)

model_D = ModelD().to(DEVICE)
model_G = ModelG().to(DEVICE)

optimizer_D = torch.optim.Adam(model_D.parameters(), lr=LEARNING_RATE, betas=(0.5, 0.9))
optimizer_G = torch.optim.Adam(model_G.parameters(), lr=LEARNING_RATE, betas=(0.5, 0.9))


# Implementation based on
# https://towardsdatascience.com/demystified-wasserstein-gan-with-gradient-penalty-ba5e9b905ead
def calc_batch_gradient_penalty(discriminator, real_data, fake_data):
    batch_size = real_data.size(0)

    # Sample interpolation coefficient from uniform distribution
    eps = torch.rand(batch_size, 1, 1, 1, device=DEVICE, requires_grad=True)
    eps = eps.expand_as(real_data)

    # Interp between real and fake distributions
    interp = eps * real_data + (1 - eps) * fake_data

    # Logits for the interpolated data
    interp_logits = discriminator(interp)
    grad_outputs = torch.ones_like(interp_logits)

    # Compute gradient of x hat
    gradients = torch.autograd.grad(
        outputs=interp_logits,
        inputs=interp,
        grad_outputs=grad_outputs,
        create_graph=True,
        retain_graph=True,
    )[0]

    gradients = gradients.view(batch_size, -1)
    grad_norm = gradients.norm(2, dim=1)
    return torch.mean((grad_norm - 1) ** 2)


metrics = {}
for stage in ['train']:
    for metric in ["loss_g", "loss_d"]:
        metrics[f'{stage}_{metric}'] = []

noise_dist = torch.distributions.Normal(loc=0., scale=1.)

critic_iters = 5

for epoch in range(1, EPOCHS):
    metrics_epoch = {key: [] for key in metrics.keys()}

    stage = 'train'
    for x, x_idx in tqdm(data_loader_train, desc=stage):
        x = x.to(DEVICE)

        # Realness -> 1, value -> -Inf
        # Realnesss -> 0, value -> +Inf

        # Train disciminator
        for _ in range(critic_iters):
            optimizer_D.zero_grad()

            # Train D on real data
            y_real = model_D.forward(x)

            # Train d on fake
            # Gen fake sample
            z_fake = noise_dist.sample((x.size(0), Z_SIZE)).to(DEVICE)
            x_fake = model_G.forward(z_fake)
            x_fake_detached = x_fake.detach()  # Detach x_fake to avoid training G from loss_D

            y_fake = model_D.forward(x_fake_detached)

            # Calc D loss
            gradient_penalty = calc_batch_gradient_penalty(model_D, x, x_fake_detached)
            loss_D = torch.mean(y_fake) - torch.mean(y_real) + PENALTY_COEF * gradient_penalty
            loss_D.backward()

            optimizer_D.step()

        optimizer_G.zero_grad()

        # Forward of fake images through D
        # fuckin cursed ass variable scoping for x_fake
        y_gen = model_D.forward(x_fake)
        loss_G = -torch.mean(y_gen)
        loss_G.backward()
        optimizer_G.step()

        metrics_epoch[f'{stage}_loss_g'].append(loss_G.cpu().item())
        metrics_epoch[f'{stage}_loss_d'].append(loss_D.cpu().item())

    metrics_strs = []
    for key in metrics_epoch.keys():
        value = 0
        if len(metrics_epoch[key]):
            value = np.mean(metrics_epoch[key])
        metrics[key].append(value)
        metrics_strs.append(f'{key}: {round(value, 2)}')

    print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.clf()
    plt.subplot(121)  # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1
    plt.legend(plts, [it.get_label() for it in plts])

    plt.subplot(122)  # row col idx
    grid_img = torchvision.utils.make_grid(
        x_fake.detach().cpu(),
        padding=10,
        scale_each=True,
        nrow=8
    )
    plt.imshow(grid_img.permute(1, 2, 0))

    plt.tight_layout(pad=0.5)

    if len(RUN_PATH) == 0:
        plt.show()
    else:
        plt.savefig(f'{RUN_PATH}/plt-{epoch}.png')