import os
import torch
import numpy as np
import torchvision
import torchvision.transforms as transforms
from torchvision.utils import save_image
from torch.autograd import Variable
import torch.nn.functional
import matplotlib.pyplot as plt
import torch.utils.data
import shutil
import torch.distributions
from tqdm import tqdm


BATCH_SIZE = 64
LEARNING_RATE_D = 3e-4
LEARNING_RATE_G = 3e-4
EMBEDDING_SIZE = 128
INPUT_SIZE = 28
IS_HEADLESS = True
RUN_NAME = 'test'
CLASS_TO_USE = 'All'
NOISE_PROP = 0.05
LABEL_FLIP = True

DEVICE = 'cpu'
if torch.cuda.is_available(): # google colab hack
    DEVICE = 'cuda'
    BATCH_SIZE = 256
    MAX_LEN = None

print(f'training on: {DEVICE}')

path_run = f'./results/{RUN_NAME}'
if os.path.exists(path_run):
    shutil.rmtree(path_run)
os.makedirs(path_run)

class DatasetMNIST(torch.utils.data.Dataset):
    def __init__(self, is_train):
        super().__init__()
        self.data = torchvision.datasets.FashionMNIST(
            root='data',
            train=is_train,
            download=True,
            transform=transforms.Compose([transforms.Resize(INPUT_SIZE), transforms.ToTensor(), transforms.Normalize([0.5], [0.5])])
        )
        if CLASS_TO_USE is not 'All':
            class_idx = self.data.class_to_idx[CLASS_TO_USE]
            self.data.data = torch.stack([self.data.data[idx] for idx in range(len(self.data.targets)) if self.data.targets[idx] == class_idx])
            self.data.targets = torch.stack([self.data.targets[idx] for idx in range(len(self.data.targets)) if self.data.targets[idx] == class_idx])

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.data)

    def __getitem__(self, idx):
        x, y_idx = self.data[idx]
        y = torch.LongTensor([y_idx])
        return x, y


data_loader_train = torch.utils.data.DataLoader(
    dataset=DatasetMNIST(is_train=True),
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=True,
)

data_loader_test = torch.utils.data.DataLoader(
    dataset=DatasetMNIST(is_train=False),
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=True
)


class ModelD(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=1, out_channels=2, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=2),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 2, 14, 14

            torch.nn.Conv2d(in_channels=2, out_channels=4, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=4),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 4, 7, 7

            torch.nn.Conv2d(in_channels=4, out_channels=8, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=8),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 8, 4, 4
            torch.nn.AdaptiveMaxPool2d((1, 1)) # B, 8, 1, 1
        )
        self.fc = torch.nn.Linear(
            in_features=8,
            out_features=1
        )

    def forward(self, x):
        out = self.encoder.forward(x)
        out = self.fc.forward(out.squeeze())
        out = torch.sigmoid(out)
        return out


class ModelD2(torch.nn.Module):
    def __init__(self):
        super().__init__()

        def discriminator_block(in_filters, out_filters, bn=True):
            block = [torch.nn.Conv2d(in_filters, out_filters, 3, 2, 1), torch.nn.LeakyReLU(0.2, inplace=True), torch.nn.Dropout2d(0.25)]
            if bn:
                block.append(torch.nn.BatchNorm2d(out_filters, 0.8))
            return block

        self.encoder = torch.nn.Sequential(
            *discriminator_block(1, 16, bn=False),
            *discriminator_block(16, 32),
            *discriminator_block(32, 64),
            *discriminator_block(64, 128),
        )

        out_size = 2
        self.fc = torch.nn.Linear(
            in_features=128 * out_size ** 2,
            out_features=1
        )

    def forward(self, x):
        out = self.encoder.forward(x)
        out = out.view(out.shape[0], -1)
        out = self.fc.forward(out)
        out = torch.sigmoid(out)
        return out

class ModelD3(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=1, out_channels=8, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=8),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 4, 14, 14

            torch.nn.Conv2d(in_channels=8, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 4, 7, 7

            torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=64),
            torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1), # B, 8, 4, 4
            torch.nn.AdaptiveMaxPool2d((1, 1)) # B, 8, 1, 1
        )
        self.fc = torch.nn.Linear(
            in_features=64,
            out_features=1
        )

    def forward(self, x):
        out = self.encoder.forward(x)
        out = self.fc.forward(out.squeeze())
        out = torch.sigmoid(out)
        return out


class ModelG(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.decoder = torch.nn.Sequential(
            torch.nn.UpsamplingNearest2d(size=(2, 2)),
            torch.nn.Conv2d(in_channels=EMBEDDING_SIZE, out_channels=8, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=8),
            torch.nn.LeakyReLU(),

            torch.nn.UpsamplingNearest2d(size=(4, 4)),
            torch.nn.Conv2d(in_channels=8, out_channels=4, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=4),
            torch.nn.LeakyReLU(),

            torch.nn.UpsamplingNearest2d(size=(8, 8)),
            torch.nn.Conv2d(in_channels=4, out_channels=4, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=4),
            torch.nn.LeakyReLU(),

            torch.nn.UpsamplingNearest2d(size=(16, 16)),
            torch.nn.Conv2d(in_channels=4, out_channels=2, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=2),
            torch.nn.LeakyReLU(),

            torch.nn.UpsamplingNearest2d(size=(INPUT_SIZE, INPUT_SIZE)), #28, 28
            torch.nn.Conv2d(in_channels=2, out_channels=1, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=1),
            torch.nn.Sigmoid()
        )

    def forward(self, z):
        z_2d = z.view(-1, EMBEDDING_SIZE, 1, 1)
        out = self.decoder.forward(z_2d)
        return out

class ModelG2(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.init_size = INPUT_SIZE // 4
        self.l1 = torch.nn.Sequential(torch.nn.Linear(EMBEDDING_SIZE, 128 * self.init_size ** 2))

        self.conv_blocks = torch.nn.Sequential(
            torch.nn.BatchNorm2d(128),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(128, 128, 3, stride=1, padding=1),
            torch.nn.BatchNorm2d(128, 0.8),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(128, 64, 3, stride=1, padding=1),
            torch.nn.BatchNorm2d(64, 0.8),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Conv2d(64, 1, 3, stride=1, padding=1),
            torch.nn.Tanh(),
        )

    def forward(self, z):
        out = self.l1(z)
        out = out.view(out.shape[0], 128, self.init_size, self.init_size)
        img = self.conv_blocks(out)
        return img


class ModelG3(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.init_size = INPUT_SIZE // 4
        self.l1 = torch.nn.Sequential(torch.nn.Linear(EMBEDDING_SIZE, 32 * self.init_size ** 2))

        self.conv_blocks = torch.nn.Sequential(
            torch.nn.BatchNorm2d(32),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(32, 16, 3, stride=1, padding=1),
            torch.nn.BatchNorm2d(16, 0.8),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(16, 8, 3, stride=1, padding=1),
            torch.nn.BatchNorm2d(8, 0.8),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Conv2d(8, 1, 3, stride=1, padding=1),
            torch.nn.Tanh(),
        )

    def forward(self, z):
        out = self.l1(z)
        out = out.view(out.shape[0], 32, self.init_size, self.init_size)
        img = self.conv_blocks(out)
        return img


def BCE(y_true, y_pred):
    return -torch.mean(y_true * torch.log(y_pred) + (1 - y_true) * torch.log(1 - y_pred + 1e-8))


def get_soft_labels(batch_size, low, high, mode):
    label = ((low - high) * torch.rand(batch_size, ) + high).to(DEVICE)
    if LABEL_FLIP and mode == 'D':
        flipped_idx = np.random.choice(np.arange(len(label)), size=int(NOISE_PROP * len(label)))
        if low == 0:
            label[flipped_idx] = 0.9 + label[flipped_idx]
        else:
            label[flipped_idx] = 1.1 - label[flipped_idx]
    return label

model_D = ModelD3()
model_D = model_D.to(DEVICE)
model_G = ModelG3()
model_G = model_G.to(DEVICE)

adversarial_loss = torch.nn.BCELoss()

total_param_size = 0
for name, param in model_G.named_parameters():
    each_param_size = np.prod(param.size())
    total_param_size += each_param_size
print(f'Generator param count: {total_param_size}')

total_param_size = 0
for name, param in model_D.named_parameters():
    each_param_size = np.prod(param.size())
    total_param_size += each_param_size
print(f'Discriminator param count: {total_param_size}')

optimizer_G = torch.optim.Adam(model_G.parameters(), lr=LEARNING_RATE_G)
optimizer_D = torch.optim.Adam(model_D.parameters(), lr=LEARNING_RATE_D)

metrics = {}
for stage in ['train']:
    for metric in [
        'loss_D',
        'loss_G',
        'loss'
    ]:
        metrics[f'{stage}_{metric}'] = []

dist = torch.distributions.Normal(
    loc=0.0,
    scale=1.0,
)
generator_counter = 0

for epoch in tqdm(range(1, 1000)):
    plt.clf()

    for data_loader in [data_loader_train]: #TODO check acc for discriminator on test set
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'

        for x, y in data_loader:

            x = x.to(DEVICE)

            if data_loader == data_loader_train:

                # Adversarial ground truths
                valid = Variable(torch.FloatTensor(x.shape[0], 1).fill_(1.0), requires_grad=False).to(DEVICE)
                fake = Variable(torch.FloatTensor(x.shape[0], 1).fill_(0.0), requires_grad=False).to(DEVICE)

                optimizer_D.zero_grad()
                # Generate fake sample
                z_fake = dist.sample((BATCH_SIZE, EMBEDDING_SIZE)).to(DEVICE)
                x_fake = model_G.forward(z_fake)

                # Train discriminator on true samples
                # label = get_soft_labels(batch_size=BATCH_SIZE, low=0.9, high=1.1, mode='D')
                y_real = model_D.forward(x)
                loss_D_real = adversarial_loss(y_real, valid)

                # Training discriminator on fake samples
                # label = get_soft_labels(batch_size=BATCH_SIZE, low=0., high=0.2, mode='D')
                y_fake = model_D.forward(x_fake.detach())
                loss_D_fake = adversarial_loss(y_fake, fake)

                coef_d_real = 0.5
                coef_d_fake = 0.5
                loss_D = coef_d_real * loss_D_real + coef_d_fake * loss_D_fake
                loss_D.backward()
                optimizer_D.step()

                # Train generator
                optimizer_G.zero_grad()
                # label = get_soft_labels(batch_size=BATCH_SIZE, low=0.9, high=1.1, mode='G')
                z_fake = dist.sample((BATCH_SIZE, EMBEDDING_SIZE)).to(DEVICE)
                x_gen = model_G.forward(z_fake)
                y_gen = model_D.forward(x_gen)

                loss_const = adversarial_loss(y_gen, valid)
                # loss_ssim = 1 - ssim(x_gen, x)
                loss_G = loss_const # + loss_ssim
                loss_G.backward()
                optimizer_G.step()

                loss = loss_D + loss_G
                metrics_epoch[f'{stage}_loss_D'].append(loss_D.cpu().item()) # Tensor(0.1) => 0.1f
                metrics_epoch[f'{stage}_loss_G'].append(loss_G.cpu().item()) # Tensor(0.1) => 0.1f
                metrics_epoch[f'{stage}_loss'].append(loss.cpu().item()) # Tensor(0.1) => 0.1f


        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.clf()
    plt.figure(figsize=(10, 5))
    plts = []
    c = 0
    for key, value in metrics.items():

        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    if IS_HEADLESS:
        plt.savefig(f'{path_run}/{epoch}_loss.png')
    else:
        plt.show()


    save_image(x_fake[:16], f'{path_run}/{epoch}_gen.png', nrow=4, normalize=True)

    plt.close('all')