import argparse # pip3 install argparse
from copy import copy

from tqdm import tqdm # pip install tqdm
import hashlib
import os
import pickle
import time
import shutil
import torch
import torch.nn.functional as F
import numpy as np
import torchvision
import random
import torch.distributed
import torch.multiprocessing as mp

import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = (15, 7)
plt.style.use('dark_background')

import torch.utils.data

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-run_path', default='', type=str)

parser.add_argument('-num_epochs', default=100, type=int)
parser.add_argument('-batch_size', default=64, type=int)
parser.add_argument('-classes_count', default=20, type=int)
parser.add_argument('-samples_per_class', default=2000, type=int)

parser.add_argument('-learning_rate', default=1e-4, type=float)
parser.add_argument('-z_size', default=128, type=int)

parser.add_argument('-coef_d', default=0.5, type=float)
parser.add_argument('-coef_g', default=0.5, type=float)

parser.add_argument('-is_debug', default=False, type=lambda x: (str(x).lower() == 'true'))

args, _ = parser.parse_known_args()

RUN_PATH = args.run_path
BATCH_SIZE = args.batch_size
EPOCHS = args.num_epochs
LEARNING_RATE = args.learning_rate
Z_SIZE = args.z_size
DEVICE = 'cuda'
MAX_LEN = args.samples_per_class
MAX_CLASSES = args.classes_count # 0 = include all
IS_DEBUG = args.is_debug
INPUT_SIZE = 28
COEF_D = args.coef_d
COEF_G = args.coef_g

if not torch.cuda.is_available() or IS_DEBUG:
    MAX_LEN = 300 # per class for debugging
    MAX_CLASSES = 6 # reduce number of classes for debugging
    DEVICE = 'cpu'
    BATCH_SIZE = 66


class DatasetEMNIST(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__() # 62 classes
        self.data = torchvision.datasets.EMNIST(
            root='../data',
            split='byclass',
            train=(MAX_LEN == 0),
            download=True
        )
        class_to_idx = self.data.class_to_idx
        idx_to_class = dict((value, key) for key, value in class_to_idx.items())
        self.labels = [idx_to_class[idx] for idx in range(len(idx_to_class))]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor
        pil_x, y_idx = self.data[idx]
        np_x = np.transpose(np.array(pil_x)).astype(np.float32)
        np_x = np.expand_dims(np_x, axis=0) / 255.0 # (1, W, H) => (1, 28, 28)
        return np_x, y_idx


class ModelD(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torchvision.models.resnet18( # 3, W, H
            pretrained=False
        )
        self.encoder.fc = torch.nn.Linear(
            in_features=self.encoder.fc.in_features,
            out_features=1
        )

    def forward(self, x):
        y_prim = self.encoder.forward(x.expand(x.size(0), 3, 28, 28))
        y_prim = torch.sigmoid(y_prim)
        return y_prim


class ModelG(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.init_size = INPUT_SIZE // 4
        self.fc = torch.nn.Linear(Z_SIZE, 128 * self.init_size ** 2)

        self.conv_blocks = torch.nn.Sequential(
            torch.nn.BatchNorm2d(128),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(128, 0.8),
            torch.nn.LeakyReLU(0.2),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv2d(128, 64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(64, 0.8),
            torch.nn.LeakyReLU(0.2),
            torch.nn.Conv2d(64, 1, kernel_size=3, stride=1, padding=1),
            torch.nn.Sigmoid(),
        )

    def forward(self, z):
        out = self.fc.forward(z)
        out = out.view(out.shape[0], 128, self.init_size, self.init_size)
        y_prim = self.conv_blocks.forward(out)
        return y_prim


class LossBCE(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, y, y_pred):
        loss = -torch.mean(y * torch.log(y_pred + 1e-8) + (1 - y) * torch.log(1 - y_pred + 1e-8))
        return loss


def setup_fpp(rank, world_size):
    os.environ['MASTER_ADDR'] = '127.0.0.1'
    os.environ['MASTER_PORT'] = '8888'
    torch.distributed.init_process_group("nccl", init_method='env://', rank=rank, world_size=world_size)

def cleanup_dpp():
    torch.distributed.destroy_process_group()

def main(DEVICE, world_size, RUN_PATH):
    print(DEVICE)

    setup_fpp(DEVICE, world_size)
    torch.cuda.set_device(DEVICE)

    dataset_full = DatasetEMNIST()

    random.seed(1)
    labels_train = copy(dataset_full.labels)
    random.shuffle(labels_train)
    labels_train = labels_train[:MAX_CLASSES]
    random.seed(int(time.time()))

    idx_train = []
    str_args_for_hasing = [str(it) for it in [MAX_LEN, MAX_CLASSES]]
    hash_args = hashlib.md5((''.join(str_args_for_hasing)).encode()).hexdigest()
    path_cache = f'../data/{hash_args}_gan.pkl'
    if os.path.exists(path_cache):
        print('loading from cache')
        with open(path_cache, 'rb') as fp:
            idx_train = pickle.load(fp)

    else:
        labels_count = dict((key, 0) for key in dataset_full.labels)
        for idx, (x, y_idx) in tqdm(
                enumerate(dataset_full),
                'splitting dataset',
                total=len(dataset_full)
        ):
            label = dataset_full.labels[y_idx]
            if MAX_LEN > 0:
                if labels_count[label] >= MAX_LEN:
                    if all(it >= MAX_LEN for it in labels_count.values()):
                        break
                    continue
            labels_count[label] += 1
            if label in labels_train:
                idx_train.append(idx)

        with open(path_cache, 'wb') as fp:
            pickle.dump(idx_train, fp)

    dataset_train = torch.utils.data.Subset(dataset_full, idx_train)
    data_loader_train = torch.utils.data.DataLoader(
        dataset=dataset_train,
        batch_size=BATCH_SIZE,
        shuffle=True,
        drop_last=(len(dataset_train) % BATCH_SIZE < 12),
        num_workers=8
    )

    model_D = ModelD().to(DEVICE)
    model_G = ModelG().to(DEVICE)
    optimizer_D = torch.optim.Adam(model_D.parameters(), lr=LEARNING_RATE)
    optimizer_G = torch.optim.Adam(model_G.parameters(), lr=LEARNING_RATE)
    loss_fn = LossBCE()

    if torch.cuda.device_count() > 1:
        model_D = torch.nn.parallel.DistributedDataParallel(model_D, device_ids=[DEVICE], output_device=DEVICE)
        model_G = torch.nn.parallel.DistributedDataParallel(model_G, device_ids=[DEVICE], output_device=DEVICE)

    dist = torch.distributions.Normal(
        loc=0.0,
        scale=1.0,
    )

    metrics = {}
    for stage in ['train']:
        for metric in ['loss', 'loss_g', 'loss_d']:
            metrics[f'{stage}_{metric}'] = []


    for epoch in range(1, 100):
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        model_D = model_D.train()
        model_G = model_G.train()
        torch.set_grad_enabled(True)

        for x, x_idx in tqdm(data_loader_train, desc=stage):
            x = x.to(DEVICE)
            #x_idx = x_idx.squeeze().to(DEVICE)

            z_fake = dist.sample((BATCH_SIZE, Z_SIZE)).to(DEVICE)
            x_fake = model_G.forward(z_fake)
            # for param in model_D.parameters():
            #     param.requires_grad = True
            y_real = model_D.forward(x)
            y_fake = model_D.forward(x_fake.detach()) # detach generated x because do not train G on loss_D

            loss_D_real = loss_fn.forward(y_real, torch.ones_like(y_real)) # 1.0 if real
            loss_D_fake = loss_fn.forward(y_fake, torch.zeros_like(y_fake))
            loss_D = COEF_D * (loss_D_real + loss_D_fake)

            torch.distributed.barrier()

            loss_D.backward()
            optimizer_D.step()
            optimizer_D.zero_grad()

            z_fake = dist.sample((BATCH_SIZE, Z_SIZE)).to(DEVICE)
            x_gen = model_G.forward(z_fake)
            # for param in model_D.parameters():
            #     param.requires_grad = False
            y_gen = model_D.forward(x_gen)
            loss_G = COEF_G * loss_fn.forward(y_gen, torch.ones_like(y_gen)) # opposite to D, fake as real

            loss = loss_D + loss_G

            torch.distributed.barrier()

            loss_G.backward()
            optimizer_G.step()
            optimizer_G.zero_grad()

            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item())
            metrics_epoch[f'{stage}_loss_g'].append(loss_G.cpu().item())
            metrics_epoch[f'{stage}_loss_d'].append(loss_D.cpu().item())

        metrics_strs = []
        for key in metrics_epoch.keys():
            value = 0
            if len(metrics_epoch[key]):
                value = np.mean(metrics_epoch[key])
            metrics[key].append(value)
            metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

        plt.clf() # BUGFIX!! scatter is remembered even when new subplot is made

        plt.subplot(121) # row col idx
        plts = []
        c = 0
        for key, value in metrics.items():
            plts += plt.plot(value, f'C{c}', label=key)
            ax = plt.twinx()
            c += 1
        plt.legend(plts, [it.get_label() for it in plts])

        plt.subplot(122)  # row col idx
        grid_img = torchvision.utils.make_grid(
            x_gen.detach().cpu(),
            padding=10,
            scale_each=True,
            nrow=8
        )
        plt.imshow(grid_img.permute(1, 2, 0))

        plt.tight_layout(pad=0.5)

        if len(RUN_PATH) == 0:
            plt.show()
        else:
            if np.isnan(metrics[f'train_loss'][-1]) or np.isinf(metrics[f'train_loss'][-1]):
                exit()
            plt.savefig(f'{RUN_PATH}/plt-{epoch}.png')

    if torch.cuda.device_count() > 1:
        cleanup_dpp()


if __name__ == "__main__":
    if len(RUN_PATH):
        RUN_PATH = f'{int(time.time())}_{RUN_PATH}'
        if os.path.exists(RUN_PATH):
            shutil.rmtree(RUN_PATH)
        os.makedirs(RUN_PATH)

    if torch.cuda.device_count() > 1:
        world_size = torch.cuda.device_count()
        mp.spawn(main,
                args=(world_size, RUN_PATH),
                nprocs=world_size,
                join=True)
    else:
        main(DEVICE, world_size=1)