# GAN hacks:
    # Use SGD for discriminator and ADAM for generator
    # Avoid Sparse Gradients: ReLU, MaxPool:
        # For Downsampling, use: Average Pooling, Conv2d + stride
        # For Upsampling, use: PixelShuffle, ConvTranspose2d + stride

import argparse # pip3 install argparse
from copy import copy

from tqdm import tqdm # pip install tqdm
import hashlib
import os
import pickle
import time
import shutil
import torch
import torch.nn.functional as F
import numpy as np
import torchvision
import random
import torch.distributed
import torch.multiprocessing as mp

from sklearn.datasets import fetch_lfw_people

import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = (15, 7)
plt.style.use('dark_background')

import torch.utils.data

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-run_path', default='', type=str)

parser.add_argument('-num_epochs', default=500, type=int)
parser.add_argument('-batch_size', default=128, type=int)
parser.add_argument('-classes_count', default=6, type=int)
parser.add_argument('-samples_per_class', default=10000, type=int)

parser.add_argument('-learning_rate', default=5e-5, type=float)
parser.add_argument('-z_size', default=128, type=int)

parser.add_argument('-is_debug', default=True, type=lambda x: (str(x).lower() == 'true'))

args, _ = parser.parse_known_args()

RUN_PATH = args.run_path
BATCH_SIZE = args.batch_size
EPOCHS = args.num_epochs
LEARNING_RATE = args.learning_rate
Z_SIZE = args.z_size
DEVICE = 'cuda'
MAX_LEN = args.samples_per_class
MAX_CLASSES = args.classes_count # 0 = include all
IS_DEBUG = False
INPUT_SIZE = 28

if not torch.cuda.is_available() or IS_DEBUG:
    MAX_LEN = 10 # per class for debugging
    MAX_CLASSES = 2 # reduce number of classes for debugging
    DEVICE = 'cpu'
    BATCH_SIZE = 32

if len(RUN_PATH):
    RUN_PATH = f'{int(time.time())}_{RUN_PATH}'
    if os.path.exists(RUN_PATH):
        shutil.rmtree(RUN_PATH)
    os.makedirs(RUN_PATH)


class DatasetLFW(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()

        lfw_people = fetch_lfw_people(min_faces_per_person=50, resize=1)
        self.img = lfw_people.images
        self.target = lfw_people.target
        self.labels = lfw_people.target_names

        # self.img = self.img[:100]
        # self.target = self.target[:100]

    def __len__(self):
        return len(self.img)

    def __getitem__(self, idx):
        pil_x = self.img[idx]
        y_idx = self.target[idx]
        # np_x = np.transpose(np.array(pil_x)).astype(np.float32)
        np_x = (np.array(pil_x)).astype(np.float32)
        np_x = np.expand_dims(np_x, axis=0) / 255.0
        return np_x, int(y_idx)


class ModelD(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=1, out_channels=8, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=8),
            torch.nn.LeakyReLU(),
            # torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),
            torch.nn.AvgPool2d(kernel_size=4, stride=2, padding=1),

            torch.nn.Conv2d(in_channels=8, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.LeakyReLU(),
            # torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),
            torch.nn.AvgPool2d(kernel_size=4, stride=2, padding=1),

            torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=64),
            torch.nn.LeakyReLU(),
            # torch.nn.MaxPool2d(kernel_size=4, stride=2, padding=1),
            torch.nn.AvgPool2d(kernel_size=4, stride=2, padding=1),

            # torch.nn.AdaptiveMaxPool2d((1, 1))
            torch.nn.AdaptiveAvgPool2d((1, 1))
        )

        self.fc = torch.nn.Linear(in_features=64, out_features=1)

    def forward(self, x):
        out = self.encoder.forward(x)
        out = self.fc.forward(out.squeeze())

        return out


class ModelG(torch.nn.Module):
    def __init__(self):
        super().__init__()

        # self.init_size = INPUT_SIZE // 4

        self.init_w = 125 // 4
        self.init_h = 94 // 4

        # self.fc = torch.nn.Linear(in_features=Z_SIZE, out_features=128 * self.init_size ** 2)
        self.fc = torch.nn.Linear(in_features=Z_SIZE, out_features=128 * self.init_w * self.init_h)

        self.decoder = torch.nn.Sequential(
            torch.nn.BatchNorm2d(num_features=128),
            # torch.nn.Upsample(scale_factor=2),
            torch.nn.PixelShuffle(upscale_factor=2),

            # torch.nn.Conv2d(in_channels=32, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=32, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=128),
            torch.nn.LeakyReLU(negative_slope=0.2),

            # torch.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=128),
            torch.nn.LeakyReLU(negative_slope=0.2),

            # torch.nn.Upsample(scale_factor=2),
            torch.nn.PixelShuffle(upscale_factor=2),

            # torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=64),
            torch.nn.LeakyReLU(negative_slope=0.2),

            # torch.nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=64),
            torch.nn.LeakyReLU(negative_slope=0.2),

            # torch.nn.Conv2d(in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1),
            torch.nn.BatchNorm2d(num_features=32),
            torch.nn.LeakyReLU(negative_slope=0.2),

            # torch.nn.Conv2d(in_channels=32, out_channels=1, kernel_size=3, stride=1, padding=1),
            torch.nn.ConvTranspose2d(in_channels=32, out_channels=1, kernel_size=3, stride=1, padding=1),

            torch.nn.Sigmoid()
        )

    def forward(self, z):
        out = self.fc.forward(z) # 32, 6272
        # out = out.view(out.shape[0], 128, self.init_size, self.init_size)
        out = out.view(out.shape[0], 128, self.init_w, self.init_h)
        out = self.decoder.forward(out)

        return out


def weights_init(m):
    class_name = m.__class__.__name__
    if class_name.find("Conv") != -1:
        torch.nn.init.normal_(m.weight.data, 0., 0.02)

    elif class_name.find("BatchNorm") != -1:
        torch.nn.init.normal_(m.weight.data, 1., 0.02)
        torch.nn.init.constant_(m.bias.data, 0)


dataset_full = DatasetLFW()

np.random.seed(2)
labels_train = copy(dataset_full.labels)
random.shuffle(labels_train, random=np.random.random)
labels_train = labels_train[:MAX_CLASSES]
np.random.seed(int(time.time()))

print(f"train labels: {labels_train}")

idx_train = []
str_args_for_hasing = [str(it) for it in [MAX_LEN, MAX_CLASSES] + labels_train.tolist()]
hash_args = hashlib.md5((''.join(str_args_for_hasing)).encode()).hexdigest()
path_cache = f'../data/{hash_args}_gan.pkl'
if os.path.exists(path_cache):
    print('loading from cache')
    with open(path_cache, 'rb') as fp:
        idx_train = pickle.load(fp)

else:
    labels_count = dict((key, 0) for key in dataset_full.labels)
    for idx, (x, y_idx) in tqdm(
            enumerate(dataset_full),
            'splitting dataset',
            total=len(dataset_full)
    ):
        label = dataset_full.labels[y_idx]
        if MAX_LEN > 0:
            if labels_count[label] >= MAX_LEN:
                if all(it >= MAX_LEN for it in labels_count.values()):
                    break
                continue
        labels_count[label] += 1
        if label in labels_train:
            idx_train.append(idx)

    with open(path_cache, 'wb') as fp:
        pickle.dump(idx_train, fp)

dataset_train = torch.utils.data.Subset(dataset_full, idx_train)
data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_full,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE < 12),
    num_workers=(8 if not IS_DEBUG else 0)
)

model_D = ModelD().to(DEVICE)
model_G = ModelG().to(DEVICE)

# optimizer_D = torch.optim.RMSprop(model_D.parameters(), lr=LEARNING_RATE)
# optimizer_G = torch.optim.RMSprop(model_G.parameters(), lr=LEARNING_RATE)

optimizer_D = torch.optim.SGD(model_D.parameters(), lr=LEARNING_RATE)
optimizer_G = torch.optim.Adam(model_G.parameters(), lr=LEARNING_RATE)

metrics = {}
for stage in ['train']:
    for metric in ['loss_g', 'loss_d']:
        metrics[f'{stage}_{metric}'] = []

dist = torch.distributions.Normal(loc=0., scale=1.)
critic_iters = 5

for epoch in range(1, EPOCHS):
    metrics_epoch = {key: [] for key in metrics.keys()}

    stage = 'train'
    for x, x_idx in tqdm(data_loader_train, desc=stage):
        x = x.to(DEVICE)

        # real should have very small value -Inf model_D/y
        # more real the more negative values

        # fake should have very large value -Inf model_D/y
        # more fake the more positive values

        # Train discriminator

        for _ in range(critic_iters):
            optimizer_D.zero_grad()

            # Train D on real data
            y_real = model_D.forward(x)


            # Train D on fake data
            # Generate fake sample

            z_fake = dist.sample((x.size(0), Z_SIZE)).to(DEVICE)
            x_fake = model_G.forward(z_fake)

            y_fake = model_D.forward(x_fake.detach()) # detach x_fake because do not train G on loss_D


            # Calculate D loss

            loss_D = torch.mean(y_fake) - torch.mean(y_real)
            loss_D.backward()

            torch.nn.utils.clip_grad_norm_(model_D.parameters(), max_norm=1e-2, norm_type=1)

            optimizer_D.step()


        # Train generator
        optimizer_G.zero_grad()

        # Forward of fake images through D
        y_gen = model_D.forward(x_fake)

        # Calculate G loss
        loss_G = -torch.mean(y_gen)
        loss_G.backward()
        optimizer_G.step()

        metrics_epoch[f'{stage}_loss_g'].append(loss_G.cpu().item())
        metrics_epoch[f'{stage}_loss_d'].append(loss_D.cpu().item())


    metrics_strs = []
    for key in metrics_epoch.keys():
        value = 0
        if len(metrics_epoch[key]):
            value = np.mean(metrics_epoch[key])
        metrics[key].append(value)
        metrics_strs.append(f'{key}: {round(value, 2)}')

    print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.clf()
    plt.subplot(121) # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        if "fake" in key or "real" in key:
            continue
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1
    plt.legend(plts, [it.get_label() for it in plts])

    plt.subplot(122)  # row col idx
    grid_img = torchvision.utils.make_grid(
        x_fake.detach().cpu(),
        padding=10,
        scale_each=True,
        nrow=8
    )
    plt.imshow(grid_img.permute(1, 2, 0))

    plt.tight_layout(pad=0.5)
    plt.pause(0.2)

    # if len(RUN_PATH) == 0:
    #     plt.show()
    # else:
    #     if np.isnan(metrics[f'train_loss'][-1]) or np.isinf(metrics[f'train_loss'][-1]):
    #         plt.savefig(f'{RUN_PATH}/plt-{epoch}.png')
    #         exit()
